﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Net.Mime;
using System.Text;

namespace FileStorage.Helper
{
    /// <summary>
    /// Helper class that exposes functionality to generate, convert or create data identifiers.
    /// </summary>
    public static class DataIdentifierHelper
    {
        /// <summary>
        /// Returns a byte array of a specified string
        /// </summary>
        /// <param name="text">The text to go into the byte array</param>
        /// <returns>A byte array of text</returns>
        private static byte[] GetBytes(string text)
        {
            return ASCIIEncoding.UTF8.GetBytes(text);
        }

        public static string ToNFileStorageOrigFileName(this Guid dataIdentifier, bool suppressNonReadableCharacters, char placeholderForUnreadableChars)
        {
            string result;

            try
            {
                byte[] bytes = dataIdentifier.ToByteArray();
                var encoding = new ASCIIEncoding();
                result = encoding.GetString(bytes);
                if (suppressNonReadableCharacters)
                {
                    //
                    // only return chars 32 - 127
                    //
                    var tempResult = new StringBuilder();
                    foreach (var c in result)
                    {
                        if (c >= 32 && c <= 127)
                        {
                            tempResult.Append(c);
                        }
                        else
                        {
                            tempResult.Append(placeholderForUnreadableChars);
                        }
                    }
                    result = tempResult.ToString();
                }
            }
            catch (Exception)
            {
                result = "****************";
            }

            return result;
        }

        public static string ToNFileStorageOrigFileName(this Guid dataIdentifier)
        {
            char placeholder = '*';
            string result = ToNFileStorageOrigFileName(dataIdentifier, true, placeholder);
            if (result.Contains(placeholder))
            {
                result = "****************";
            }
            return result;
        }

        public static Guid ToNFileStorageDataIdentifier(this string text)
        {
            return ToNFileStorageDataIdentifier(text, true, true, true, '_');
        }

        public static Guid ToNFileStorageDataIdentifier(this string text, bool autoPadLeft, bool autoTakeFirst16Chars, bool caseInsensitive, char padLeftChar)
        {
            if (autoPadLeft)
            {
                text = text.PadLeft(16, padLeftChar);
            }
            if (caseInsensitive)
            {
                text = text.ToLowerInvariant();
            }

            byte[] bytes = GetBytes(text);
            if (autoTakeFirst16Chars)
            {
                bytes = bytes.Take(16).ToArray();
            }

            int length = bytes.Length;

            if (length > 16)
            {
                throw new Exception(string.Format("The text {0} you provided cannot be used for a dataIdentifier; its too big (resulting in {1} bytes)", text, length));
            }
            if (length < 16)
            {
                throw new Exception(string.Format("The text {0} you provided cannot be used for a dataIdentifier; its too small (resulting in {1} bytes). Consider appending spaces?", text, length));
            }

            return new Guid(bytes);
        }

        /// <summary>
        /// Returns a GUID based upon a datetime utc format in ticks
        /// </summary>
        public static Guid CreateDataIdentifierByDateTimeUtcTimeTicks()
        {
            Guid result;

            result = DateTime.UtcNow.Ticks.ToString().ToNFileStorageDataIdentifier();

            return result;
        }

        /// <summary>
        /// Returns a GUID based upon a datetime utc format (yyyymmddhhmmssmmmm)
        /// </summary>
        public static Guid CreateDataIdentifierByReadableDateTimeUtc()
        {
            Guid result;

            var now = DateTime.UtcNow;

            var timestamp = new StringBuilder();

            var year = now.Year.ToString().PadLeft(4, '0');
            timestamp.Append(year);

            var month = now.Month.ToString().PadLeft(2, '0');
            timestamp.Append(month);

            var day = now.Day.ToString().PadLeft(2, '0');
            timestamp.Append(day);

            var hour = now.Hour.ToString().PadLeft(2, '0');
            timestamp.Append(hour);

            var minute = now.Minute.ToString().PadLeft(2, '0');
            timestamp.Append(minute);

            var secs = now.Second.ToString().PadLeft(2, '0');
            timestamp.Append(secs);

            var msecs = now.Millisecond.ToString().PadLeft(4, '0');
            timestamp.Append(msecs);

            result = timestamp.ToString().ToNFileStorageDataIdentifier();

            return result;
        }
    }
}
